/*
 * PRINTF.C
 *
 * Implementation for the core printf engine.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN 1
#endif

#include <windows.h>
#include <tchar.h>

#define MINICRT_BUILD
#include "minicrt.h"

#ifdef PRINTF_FN
#undef PRINTF_FN
#undef PRINTF_DESTLENGTH
#undef PRINTF_PUSHCHAR
#endif


#ifdef PRINTF_SIZEONLY

#ifdef UNICODE
#define PRINTF_FN mini_vswprintf_size
#else
#define PRINTF_FN mini_vsprintf_size
#endif

#define PRINTF_DESTLENGTH() (1)
#define PRINTF_PUSHCHAR(x)  dest_offset++,x;

#else // PRINTF_SIZEONLY

#ifdef UNICODE
#define PRINTF_FN mini_vswprintf_s
#else
#define PRINTF_FN mini_vsprintf_s
#endif

#define PRINTF_DESTLENGTH()  (dest_offset < len - 1)
#define PRINTF_PUSHCHAR(x)   szDest[dest_offset++] = x;

#endif // PRINTF_SIZEONLY

int
PRINTF_FN(
#ifndef PRINTF_SIZEONLY
        LPTSTR szDest,
        unsigned int len,
#endif
        LPCTSTR szFmt,
        va_list marker)
{
    unsigned int dest_offset = 0;
    unsigned int src_offset = 0;
    unsigned int i;

    BOOL leadingzero;
    BOOL leftalign;
    unsigned int element_len;

    while (szFmt[src_offset] != '\0' && PRINTF_DESTLENGTH()) {
        if (szFmt[src_offset] == '%') {
            src_offset++;
            leadingzero = FALSE;
            leftalign = FALSE;
            element_len = 0;

            if (szFmt[src_offset] == '-') {
                leftalign = TRUE;
                src_offset++;
            }
            if (szFmt[src_offset] == '0') {
                leadingzero = TRUE;
                src_offset++;
            }
            while (szFmt[src_offset] >= '0' && szFmt[src_offset] <= '9') {
                element_len = element_len * 10 + szFmt[src_offset] - '0';
                src_offset++;
            }

            if (element_len == 0) {
                element_len = (unsigned int) -1;
            }

            switch(szFmt[src_offset]) {
                case '%':
                    PRINTF_PUSHCHAR('%');
                    break;

                case 'c':
                    {
                        i = va_arg(marker, TCHAR);
                        PRINTF_PUSHCHAR(i);
                    }
                    break;
                case 's':
                    {
                        LPTSTR str = va_arg(marker, LPTSTR);

                        if (str == NULL) {
                            str = _T("(null)");
                        }

                        if (element_len != (unsigned int)-1 && !leftalign) {
                            unsigned int str_len = 0;
                            while (str[str_len] != '\0') str_len++;

                            while (element_len > str_len && PRINTF_DESTLENGTH()) {
                                PRINTF_PUSHCHAR(' ');
                                element_len--;
                            }
                        }

                        while (*str != '\0' && PRINTF_DESTLENGTH() && element_len) {
                            PRINTF_PUSHCHAR(*str);
                            str++;
                            element_len--;
                        }

                        if (leftalign) {
                            while (element_len-- && PRINTF_DESTLENGTH()) {
                                PRINTF_PUSHCHAR(' ');
                            }
                        }
                    }
                    break;
                case 'd':
                case 'i':
                case 'x':
                    {
                        unsigned int num, tempnum;
                        unsigned int divisor, digits;
                        unsigned int radix = 10;

                        //
                        //  If we're %i we're base 10, if we're %x we're
                        //  base 16
                        //

                        if (szFmt[src_offset] == 'x') {
                            radix = 16;
                        }

                        num = va_arg(marker, int);

                        //
                        //  Count the number of digits we have in the user's
                        //  input.  Stop if we hit the format specifier.
                        //  Code below will preserve low order values.
                        //

                        divisor = 1;
                        digits = 1;
                        tempnum = num;
                        while (tempnum > radix - 1 && digits < element_len) {
                            divisor *= radix;
                            tempnum = num / divisor;
                            digits++;
                        }

                        //
                        //  If the field specifier is larger, pad it with
                        //  either a zero or space depending on the format
                        //

                        if (element_len != (unsigned int)-1 && digits < element_len) {
                            tempnum = element_len - digits;
                            while (tempnum > 0) {
                                if (leadingzero) {
                                    PRINTF_PUSHCHAR('0');
                                } else {
                                    PRINTF_PUSHCHAR(' ');
                                }
                                if (!PRINTF_DESTLENGTH()) {
                                    break;
                                }
                                tempnum--;
                            }
                        }

                        do {
                            tempnum = (num / divisor) % radix;

                            if (tempnum > 9) {
                                PRINTF_PUSHCHAR((UCHAR)tempnum + 'a' - 10);
                            } else {
                                PRINTF_PUSHCHAR((UCHAR)tempnum + '0');
                            }

                            if (!PRINTF_DESTLENGTH()) {
                                break;
                            }

                            divisor /= radix;
                            digits--;

                        } while(digits > 0);

                    }
                    break;
                default:
                    {
                        LPTSTR szErr = _T("FMTERR");
                        i = 0;
                        while (PRINTF_DESTLENGTH() && szErr[i] != '\0') {
                            PRINTF_PUSHCHAR(szErr[i++]);
                        }
                    }
                    break;
            }

            src_offset++;

        } else {
            PRINTF_PUSHCHAR(szFmt[src_offset++]);
        }
    }

#ifndef PRINTF_SIZEONLY
    if (!PRINTF_DESTLENGTH()) {
        szDest[0] = '\0';
        return -1;
    }
#endif
    PRINTF_PUSHCHAR('\0');

    return dest_offset;
}

// vim:sw=4:ts=4:et:
