/*
 * STRING.C
 *
 * Implementations of the str* library functions.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN 1
#endif

#include <windows.h>
#include <tchar.h>

#define MINICRT_BUILD
#include "minicrt.h"


#ifdef UNICODE
#define mini_ttoi mini_wtoi
#else
#define mini_ttoi mini_atoi
#endif

int mini_ttoi (TCHAR * str)
{
    int ret = 0;
    while (*str >= '0' && *str <= '9') {
        ret *= 10;
        ret += *str - '0';
        str++;
    }
    return ret;
}

#ifdef UNICODE
#define mini_tcscat_s mini_wcscat_s
#else
#define mini_tcscat_s mini_strcat_s
#endif

TCHAR * mini_tcscat_s(TCHAR * dest, unsigned int len, const TCHAR * src)
{
    unsigned int i,j;
    for (i = 0; dest[i] != '\0' && i < len; i++);
    for (j = 0; src[j] != '\0' && i < len - 1; ) {
        dest[i++] = src[j++];
    }
    dest[i++] = '\0';
    return dest;
}

#ifdef UNICODE
#define mini_tcsncat mini_wcsncat
#else
#define mini_tcsncat mini_strncat
#endif

TCHAR * mini_tcsncat(TCHAR * dest, const TCHAR * src, unsigned int len)
{
    unsigned int i,j;
    for (i = 0; dest[i] != '\0'; i++);
    for (j = 0; src[j] != '\0' && j < len; ) {
        dest[i++] = src[j++];
    }
    dest[i++] = '\0';
    return dest;
}

#ifdef UNICODE
#define mini_tcschr mini_wcschr
#else
#define mini_tcschr mini_strchr
#endif

TCHAR * mini_tcschr(const TCHAR * str, TCHAR ch)
{
    const TCHAR * ptr = str;
    while (*ptr != '\0' && *ptr != ch) ptr++;
    if (*ptr == ch) return (TCHAR *)ptr;
    return NULL;
}

#ifdef UNICODE
#define mini_tcsrchr mini_wcsrchr
#else
#define mini_tcsrchr mini_strrchr
#endif


TCHAR * mini_tcsrchr(const TCHAR * str, TCHAR ch)
{
    const TCHAR * ptr = str;
    while (*ptr != '\0') ptr++;
    while (*ptr != ch && ptr > str) ptr--;
    if (*ptr == ch) return (TCHAR *)ptr;
    return NULL;
}

#ifdef UNICODE
#define mini_tcstok_s mini_wcstok_s
#else
#define mini_tcstok_s mini_strtok_s
#endif

TCHAR * mini_tcstok_s(TCHAR * str, TCHAR * match, TCHAR ** context)
{
    TCHAR * next;
    if (str != NULL) {
        *context = str;
    }

    next = *context;
    if (next == NULL) {
        return NULL;
    }

    while (*next != match[0] && *next != '\0') next++;

    if (*next == match[0]) {
        TCHAR * ret = *context;
        *next = '\0';
        *context = next++;
        return ret;
    } else {
        TCHAR * ret = *context;
        *context = NULL;
        return ret;
    }
}

#ifdef UNICODE
#define mini_tcstok mini_wcstok
#define STRTOK_CTX  wcstok_context
#else
#define mini_tcstok mini_strtok
#define STRTOK_CTX  strtok_context
#endif

TCHAR * STRTOK_CTX;
TCHAR * mini_tcstok(TCHAR * str, TCHAR * match)
{
    return mini_tcstok_s(str, match, &STRTOK_CTX);
}

#ifdef UNICODE
#define mini_stprintf_s mini_swprintf_s
#else
#define mini_stprintf_s mini_sprintf_s
#endif

int
mini_stprintf_s(LPTSTR szDest, unsigned int len, LPCTSTR szFmt, ...)
{
    va_list marker;
    int out_len;

    va_start( marker, szFmt );
    out_len = mini_vstprintf_s(szDest, len, szFmt, marker);
    va_end( marker );
    return out_len;
}

#ifdef UNICODE
#define mini_stprintf_s mini_swprintf_s
#else
#define mini_stprintf_s mini_sprintf_s
#endif

int
mini_stprintf(LPTSTR szDest, LPCTSTR szFmt, ...)
{
    va_list marker;
    int out_len;

    va_start( marker, szFmt );
    out_len = mini_vstprintf_s(szDest, (DWORD)-1, szFmt, marker);
    va_end( marker );
    return out_len;
}


#ifdef UNICODE
#define mini_vftprintf mini_vfwprintf
#else
#define mini_vftprintf mini_vfprintf
#endif

int
mini_vftprintf(int fp, LPCTSTR szFmt, va_list marker)
{
    va_list savedmarker = marker;
    int len;
    TCHAR stack_buf[20];
    HANDLE hOut;
    DWORD written;
    TCHAR * buf;

    if (fp==stdout) {
        hOut = GetStdHandle(STD_OUTPUT_HANDLE);
    } else if (fp==stderr) {
        hOut = GetStdHandle(STD_ERROR_HANDLE);
    } else {
        return 0;
    }

    len = mini_vstprintf_size(szFmt, marker);

    if (len>sizeof(buf)/sizeof(buf[0])) {
        buf = HeapAlloc(GetProcessHeap(), 0, len * sizeof(TCHAR));
        if (buf == NULL) {
            return 0;
        }
    } else {
        buf = stack_buf;
    }

    marker = savedmarker;
    len = mini_vstprintf_s(buf, len, szFmt, marker);

    WriteFile(hOut,buf,(len - 1)*sizeof(TCHAR),&written,NULL);

    if (buf != stack_buf) {
        HeapFree(GetProcessHeap(), 0, buf);
    }
    return len;
}

#ifdef UNICODE
#define mini_ftprintf mini_fwprintf
#else
#define mini_ftprintf mini_fprintf
#endif

int
mini_ftprintf(int fp, LPCTSTR szFmt, ...)
{
    va_list marker;
    int out_len;

    va_start( marker, szFmt );
    out_len = mini_vftprintf(fp, szFmt, marker);
    va_end( marker );
    return out_len;
}

#ifdef UNICODE
#define mini_tprintf mini_wprintf
#else
#define mini_tprintf mini_printf
#endif

int
mini_tprintf(LPCTSTR szFmt, ...)
{
    va_list marker;
    int out_len;

    va_start( marker, szFmt );
    out_len = mini_vftprintf(stdout, szFmt, marker);
    va_end( marker );
    return out_len;
}


// vim:sw=4:ts=4:et:
