/*
 * SDIR.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN 1
#endif

#include <windows.h>
#include <winioctl.h>
#ifndef MINICRT
#include <stdio.h>
#include <stdlib.h>
#endif
#include <tchar.h>

#ifdef MINICRT
#include <minicrt.h>
#endif


//
//  Compile time configuration
//

#define SDIR_MAX_DIRENTS 10000
#define SDIR_MAX_WIDTH   500

typedef struct _SDIR_EXTENSION_MAPPING {
    TCHAR * Extension;
    WORD Attributes;
} SDIR_EXTENSION_MAPPING, *PSDIR_EXTENSION_MAPPING;

SDIR_EXTENSION_MAPPING SdirExtensionMappings[] = {
    {_T("BAT"),            FOREGROUND_RED                                      | FOREGROUND_INTENSITY },
    {_T("CMD"),            FOREGROUND_RED                                      | FOREGROUND_INTENSITY },
    {_T("COM"),                             FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("DLL"),                             FOREGROUND_BLUE | FOREGROUND_GREEN                        },
    {_T("DOC"),            FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("DOCX"),           FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("EXE"),                             FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("HTM"),            FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("HTML"),           FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("PDF"),            FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("PL"),             FOREGROUND_RED                                      | FOREGROUND_INTENSITY },
    {_T("PPT"),            FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("PPTX"),           FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("SYS"),                             FOREGROUND_BLUE | FOREGROUND_GREEN                        },
    {_T("XLS"),            FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
    {_T("XLSX"),           FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY },
};

#define SDIR_GRID_COLOR           (FOREGROUND_GREEN)
#define SDIR_FILE_SIZE_COLOR      (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY)
#define SDIR_FILE_TIME_COLOR      (FOREGROUND_RED|FOREGROUND_GREEN)
#define SDIR_DEFAULT_COLOR        (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE)
#define SDIR_DIR_COLOR            (FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_INTENSITY)
#define SDIR_LNK_COLOR            (FOREGROUND_RED|FOREGROUND_BLUE)
#define SDIR_FUTURE_VERSION_COLOR (FOREGROUND_RED|FOREGROUND_INTENSITY)

//
//  Hacks for older compilers
//

#if (defined(_MSC_VER) && _MSC_VER >= 1200 && _MSC_VER < 1310)
#pragma comment(linker, "/OPT:NOWIN98")
#endif

#ifdef _ULONGLONG_
#define SDIR_BIG_SIZE 1
typedef ULONGLONG SDIR_FILESIZE;
#elif !defined(_MSC_VER) || _MSC_VER >= 900
#define SDIR_BIG_SIZE 1
typedef LONGLONG SDIR_FILESIZE;
#else
#pragma message("64 bit file size support not present")
typedef ULONG SDIR_FILESIZE;
#endif

#ifndef FILE_SHARE_DELETE
#define FILE_SHARE_DELETE 4
#endif

#ifndef FILE_ATTRIBUTE_SPARSE_FILE
#define FILE_ATTRIBUTE_SPARSE_FILE        (0x200)
#endif

#ifndef FILE_ATTRIBUTE_REPARSE_POINT
#define FILE_ATTRIBUTE_REPARSE_POINT      (0x400)
#endif

#ifndef FILE_ATTRIBUTE_COMPRESSED
#define FILE_ATTRIBUTE_COMPRESSED         (0x800)
#endif

#ifndef FILE_ATTRIBUTE_OFFLINE
#define FILE_ATTRIBUTE_OFFLINE           (0x1000)
#endif

#ifndef FILE_ATTRIBUTE_ENCRYPTED
#define FILE_ATTRIBUTE_ENCRYPTED         (0x4000)
#endif

#ifndef FILE_ATTRIBUTE_INTEGRITY_STREAM
#define FILE_ATTRIBUTE_INTEGRITY_STREAM  (0x8000)
#endif

#ifndef FSCTL_READ_FILE_USN_DATA
#define FSCTL_READ_FILE_USN_DATA         CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 58,  METHOD_NEITHER, FILE_ANY_ACCESS)
#endif

#ifndef _stprintf_s
#define _stprintf_s _sntprintf
#endif

#ifndef _tcstok_s
#define _tcstok_s(a,b,c) _tcstok(a,b)
#endif

#ifndef _ttoi
#ifdef UNICODE
#define _ttoi _wtoi
#else
#define _ttoi atoi
#endif
#endif

#ifndef __in
#define __in
#define __out
#define __inout
#define __in_ecount(x)
#define __out_ecount(x)
#pragma warning(disable: 4068) // unknown pragma
#endif

typedef struct {

    DWORD          RecordLength;
    WORD           MajorVersion;
    WORD           MinorVersion;
    DWORDLONG      FileReferenceNumber;
    DWORDLONG      ParentFileReferenceNumber;
    LARGE_INTEGER  Usn;
    LARGE_INTEGER  TimeStamp;
    DWORD          Reason;
    DWORD          SourceInfo;
    DWORD          SecurityId;
    DWORD          FileAttributes;
    WORD           FileNameLength;
    WORD           FileNameOffset;
    WCHAR          FileName[1];

} SDIR_USN_RECORD, *PSDIR_USN_RECORD;

typedef struct _SDIR_WIN32_FIND_STREAM_DATA {
    LARGE_INTEGER StreamSize;
    WCHAR cStreamName[MAX_PATH + 36];
} SDIR_WIN32_FIND_STREAM_DATA, *PSDIR_WIN32_FIND_STREAM_DATA;

typedef struct _SDIR_ATTRPAIR {
    DWORD FileAttribute;
    TCHAR DisplayLetter;
} SDIR_ATTRPAIR, *PSDIR_ATTRPAIR;

SDIR_ATTRPAIR AttrPairs[] = {
    {FILE_ATTRIBUTE_ARCHIVE,           'A'},
    {FILE_ATTRIBUTE_READONLY,          'R'},
    {FILE_ATTRIBUTE_HIDDEN,            'H'},
    {FILE_ATTRIBUTE_SYSTEM,            'S'},
    {FILE_ATTRIBUTE_COMPRESSED,        'C'},
    {FILE_ATTRIBUTE_ENCRYPTED,         'E'},
    {FILE_ATTRIBUTE_OFFLINE,           'O'},
    {FILE_ATTRIBUTE_REPARSE_POINT,     'r'},
    {FILE_ATTRIBUTE_SPARSE_FILE,       's'},
    {FILE_ATTRIBUTE_INTEGRITY_STREAM,  'I'},
    };

//
//  Run time configuration data
//

typedef struct _SDIR_FMTCHAR {
    TCHAR Char;
    WORD  Attr;
} SDIR_FMTCHAR, *PSDIR_FMTCHAR;

typedef struct _SDIR_DIRENT {
    LARGE_INTEGER FileId;
    LARGE_INTEGER FileSize;
    LARGE_INTEGER CompressedFileSize;
    LARGE_INTEGER Usn;
    LARGE_INTEGER FileVersion;
    DWORD         FileVersionFlags;
    WORD          OsVersionHigh;
    WORD          OsVersionLow;
    SYSTEMTIME    CreateTime;
    SYSTEMTIME    WriteTime;
    DWORD         LinkCount;
    DWORD         StreamCount;
    DWORD         ReparseTag;
    DWORD         FileAttributes;
    DWORD         FileNameLengthInChars;
    TCHAR         ShortFileName[14];
    TCHAR         FileName[MAX_PATH];
    TCHAR *       Extension;
} SDIR_DIRENT, *PSDIR_DIRENT;

typedef BOOL (WINAPI * DISABLE_WOW_REDIRECT_FN)(PVOID*);
typedef HANDLE (WINAPI * FIND_FIRST_STREAM_FN)(LPCTSTR, DWORD, PSDIR_WIN32_FIND_STREAM_DATA, DWORD);
typedef BOOL (WINAPI * FIND_NEXT_STREAM_FN)(HANDLE, PSDIR_WIN32_FIND_STREAM_DATA);
typedef DWORD (WINAPI * GET_COMPRESSED_FILE_SIZE_FN)(LPCTSTR, LPDWORD);
typedef BOOL (WINAPI * GET_DISK_FREE_SPACE_EX_FN)(LPCTSTR, PLARGE_INTEGER, PLARGE_INTEGER, PLARGE_INTEGER);

typedef DWORD (* SDIR_METADATA_WIDTH_FN)(PSDIR_FMTCHAR, WORD, PSDIR_DIRENT);
typedef DWORD (* SDIR_COMPARE_FN)(PSDIR_DIRENT, PSDIR_DIRENT);
typedef DWORD (* SDIR_COLLECT_FN)(PSDIR_DIRENT, PWIN32_FIND_DATA, LPTSTR);

#define SDIR_OPT_COLLECT (0x01)
#define SDIR_OPT_DISPLAY (0x02)

typedef struct _SDIR_OPTS {
    TCHAR           SubDirWalk[MAX_PATH];

    TCHAR           ParentName[MAX_PATH];
    DWORD           ParentNameLength;

    DWORD           OsVersion;
    DWORD           ConsoleHeight;
    DWORD           ConsoleWidth;
    DWORD           MetadataWidth;

    UCHAR           EnableCompressedFileSize;
    UCHAR           EnableCreateDate;
    UCHAR           EnableCreateTime;
    UCHAR           EnableFileAttributes;
    UCHAR           EnableFileId;
    UCHAR           EnableFileName;
    UCHAR           EnableFileSize;
    UCHAR           EnableLinkCount;
    UCHAR           EnableOsVersion;
    UCHAR           EnableReparseTag;
    UCHAR           EnableShortName;
    UCHAR           EnableStreamCount;
    UCHAR           EnableSummary;
    UCHAR           EnableUsn;
    UCHAR           EnableVersion;
    UCHAR           EnableWriteDate;
    UCHAR           EnableWriteTime;

    BOOLEAN         EnablePause;
    BOOLEAN         EnableNameTruncation;
    BOOLEAN         OutputIsConsole;
    BOOLEAN         OutputIsUnicode;
    BOOLEAN         Cancelled;

    LPTSTR          NewLine;

    SDIR_COMPARE_FN CompareFn;
    DWORD           CompareBreakCondition;
    DWORD           CompareInverseCondition;

    FIND_FIRST_STREAM_FN        FindFirstStreamW;
    FIND_NEXT_STREAM_FN         FindNextStreamW;
    GET_COMPRESSED_FILE_SIZE_FN GetCompressedFileSize;
    GET_DISK_FREE_SPACE_EX_FN   GetDiskFreeSpaceEx;

} SDIR_OPTS, *PSDIR_OPTS;

typedef struct _SDIR_SUMMARY {
    ULONG           NumFiles;
    ULONG           NumDirs;
    SDIR_FILESIZE   TotalSize;
    SDIR_FILESIZE   CompressedSize;
    LARGE_INTEGER   FreeSize;
    LARGE_INTEGER   VolumeSize;
} SDIR_SUMMARY, *PSDIR_SUMMARY;

typedef struct _SDIR_OPT {
    PUCHAR                 Enable;
    TCHAR                  Switch[3];
    UCHAR                  Default;
    SDIR_METADATA_WIDTH_FN WidthFn;
    SDIR_COLLECT_FN        CollectFn;
    SDIR_COMPARE_FN        CompareFn;
    TCHAR                  Help[20];
} SDIR_OPT, *PSDIR_OPT;

PSDIR_DIRENT SdirDirCollection = NULL;
PSDIR_DIRENT * SdirDirSorted   = NULL;
ULONG SdirDirCollectionCurrent = 0;
ULONG SdirDirCollectionLongest = 0;
ULONG SdirDirCollectionTotalNameLength = 0;

SDIR_OPTS Opts;
SDIR_SUMMARY Summary;

//
//  Display support
//

BOOL
SdirWriteStringToFile (
    __in HANDLE hFile,
    __in_ecount(count) LPCTSTR str,
    __in DWORD count
    )
{
    DWORD written;

#ifdef UNICODE
    if (!Opts.OutputIsUnicode) {
        CHAR AsciiStackBuffer[64 + 1];
        LPSTR StringBuffer;
    
        if (count < sizeof(AsciiStackBuffer)) {
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                AsciiStackBuffer,
                                sizeof(AsciiStackBuffer),
                                NULL,
                                NULL);
        
            return WriteFile(hFile, AsciiStackBuffer, count, &written, NULL);
    
        } else {
            BOOL ret;
    
            StringBuffer = HeapAlloc(GetProcessHeap(), 0, count + 1);
    
            //
            //  We'd normally display an error here, but we're probably here
            //  because we're trying to display an error - that's where large
            //  strings can arrive here.
            //
    
            if (StringBuffer == NULL) {
                return FALSE;
            }
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                StringBuffer,
                                count + 1,
                                NULL,
                                NULL);
        
            ret = WriteFile(hFile, StringBuffer, count, &written, NULL);
    
            HeapFree(GetProcessHeap(), 0, StringBuffer);
    
            return ret;
        }
    } else {
        return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
    }
#else
    return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
#endif
}


BOOL
SdirWrite (
    __in_ecount(count) PSDIR_FMTCHAR str,
    __in DWORD count
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;
    DWORD i, j;

    WORD CacheAttr;
    TCHAR CharCache[64];

    //
    //  We firstly load as much as we can into a stack buffer, then write it
    //  out.  Those syscalls are expensive - run in a slow VM to see.  We
    //  have to write out the buffer either when the next character has 
    //  different color (so we can't use the same WriteConsole call) or when
    //  the stack buffer is full.
    //

    j = 0;
    if (Opts.OutputIsConsole) {
    
        for (i = 0; i < count; i++) { 
    
            CacheAttr = str[i].Attr;
            CharCache[j] = str[i].Char;
    
            j++;
    
            if (( i + 1 < count && str[i + 1].Attr != CacheAttr) || (j >= sizeof(CharCache)/sizeof(CharCache[0]))) {
    
                SetConsoleTextAttribute(hConsole, CacheAttr);
                WriteConsole(hConsole, CharCache, j, &written, NULL);
                j = 0;
            }
        }

    } else {

        for (i = 0; i < count; i++) { 
    
            CharCache[j] = str[i].Char;
            j++;
    
            if (j >= sizeof(CharCache)/sizeof(CharCache[0])) {

                SdirWriteStringToFile(hConsole, CharCache, j);
                j = 0;
            }
        }
    }

    //
    //  If we have anything left, flush it now.
    //

    if (j > 0) {

        if (Opts.OutputIsConsole) {
            SetConsoleTextAttribute(hConsole, CacheAttr);
            WriteConsole(hConsole, CharCache, j, &written, NULL);
        } else {
            SdirWriteStringToFile(hConsole, CharCache, j);
        }
    }

    return TRUE;
}

BOOL
SdirWriteStringWithAttribute (
    __in LPCTSTR str,
    __in WORD DefaultAttribute
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;

    if (Opts.OutputIsConsole) {

        SetConsoleTextAttribute(hConsole, DefaultAttribute);
        WriteConsole(hConsole, str, (DWORD)_tcslen(str), &written, NULL);

    } else {

        SdirWriteStringToFile(hConsole, str, (DWORD)_tcslen(str));
    }

    return TRUE;
}

#define SdirWriteString(str) SdirWriteStringWithAttribute(str, SDIR_DEFAULT_COLOR)

BOOL
SdirPasteStrAndPad (
    __out_ecount(padsize) PSDIR_FMTCHAR str,
    __in_ecount(count) LPTSTR src,
    __in WORD attr,
    __in DWORD count,
    __in DWORD padsize
    )
{
    ULONG i;

    for (i = 0; i < count && i < padsize; i++) {
        str[i].Char = src[i];
        str[i].Attr = attr;
    }

    for (; i < padsize; i++) {
        str[i].Char = ' ';
        str[i].Attr = SDIR_DEFAULT_COLOR;
    }

    return TRUE;
}

#define SdirPasteStr(str, src, attr, count) \
    SdirPasteStrAndPad(str, src, attr, count, count)

BOOL
SdirDisplayError (
    __in LONG ErrorCode,
    __in LPCTSTR Prefix
    )
{
    LPTSTR OutputBuffer;

    if (FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL,
            ErrorCode,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR)&OutputBuffer,
            0,
            NULL)) {

        if (Prefix) {
            SdirWriteString(Prefix);
            SdirWriteString(_T(": "));
        }
        SdirWriteString(OutputBuffer);
        LocalFree((PVOID)OutputBuffer);
    }
    return TRUE;
}

BOOL
SdirPressAnyKey ()
{
    HANDLE hConsole = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD InputBuffer;
    DWORD NumRead;

    SdirWriteString(_T("Press any key to continue..."));

    //
    //  Loop throwing away events until we get a key pressed
    //

    while (ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0 &&
           (InputBuffer.EventType != KEY_EVENT ||
            !InputBuffer.Event.KeyEvent.bKeyDown ||
            InputBuffer.Event.KeyEvent.uChar.UnicodeChar == '\0'));

    if (InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'q' ||
        InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'Q') {

        return FALSE;
    }

    //
    //  Keep throwing away any remaining events so we're not left
    //  with stale events next time we run
    //

    while (PeekConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0) {

        ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead);
    }

    SdirWriteString(_T("\n"));
    return TRUE;
}

//
//  Sorting support
//

#define SDIR_LESS_THAN    0
#define SDIR_EQUAL        1
#define SDIR_GREATER_THAN 2

DWORD
SdirCompareLargeInt (
    __in PULARGE_INTEGER Left,
    __in PULARGE_INTEGER Right
    )
{
    if (Left->HighPart < Right->HighPart) {
        return SDIR_LESS_THAN;
    } else if (Left->HighPart > Right->HighPart) {
        return SDIR_GREATER_THAN;
    }

    if (Left->LowPart < Right->LowPart) {
        return SDIR_LESS_THAN;
    } else if (Left->LowPart > Right->LowPart) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareString (
     __in LPCTSTR Left,
     __in LPCTSTR Right
     )
{
    int ret = _tcsicmp(Left, Right);

    if (ret < 0) {
        return SDIR_LESS_THAN;
    } else if (ret == 0) {
        return SDIR_EQUAL;
    } else {
        return SDIR_GREATER_THAN;
    }
}

DWORD
SdirCompareTime (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    )
{
    if (Left->wYear < Right->wYear) {
        return SDIR_LESS_THAN;
    } else if (Left->wYear > Right->wYear) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMonth < Right->wMonth) {
        return SDIR_LESS_THAN;
    } else if (Left->wMonth > Right->wMonth) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wDay < Right->wDay) {
        return SDIR_LESS_THAN;
    } else if (Left->wDay > Right->wDay) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wHour < Right->wHour) {
        return SDIR_LESS_THAN;
    } else if (Left->wHour > Right->wHour) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMinute < Right->wMinute) {
        return SDIR_LESS_THAN;
    } else if (Left->wMinute > Right->wMinute) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wSecond < Right->wSecond) {
        return SDIR_LESS_THAN;
    } else if (Left->wSecond > Right->wSecond) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMilliseconds < Right->wMilliseconds) {
        return SDIR_LESS_THAN;
    } else if (Left->wMilliseconds > Right->wMilliseconds) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareCompressedFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->CompressedFileSize, (PULARGE_INTEGER)&Right->CompressedFileSize);
}

DWORD
SdirCompareCreateTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->CreateTime, &Right->CreateTime);
}

DWORD
SdirCompareFileId (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileId, (PULARGE_INTEGER)&Right->FileId);
}

DWORD
SdirCompareFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->FileName, Right->FileName);
}

DWORD
SdirCompareFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileSize, (PULARGE_INTEGER)&Right->FileSize);
}

DWORD
SdirCompareLinkCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->LinkCount < Right->LinkCount) {
        return SDIR_LESS_THAN;
    } else if (Left->LinkCount > Right->LinkCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareOsVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->OsVersionHigh < Right->OsVersionHigh) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionHigh > Right->OsVersionHigh) {
        return SDIR_GREATER_THAN;
    }

    if (Left->OsVersionLow < Right->OsVersionLow) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionLow > Right->OsVersionLow) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareReparseTag (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->ReparseTag < Right->ReparseTag) {
        return SDIR_LESS_THAN;
    } else if (Left->ReparseTag > Right->ReparseTag) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareShortName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->ShortFileName, Right->ShortFileName);
}

DWORD
SdirCompareStreamCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->StreamCount < Right->StreamCount) {
        return SDIR_LESS_THAN;
    } else if (Left->StreamCount > Right->StreamCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareUsn (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->Usn, (PULARGE_INTEGER)&Right->Usn);
}

DWORD
SdirCompareVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileVersion, (PULARGE_INTEGER)&Right->FileVersion);
}

DWORD
SdirCompareWriteTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->WriteTime, &Right->WriteTime);
}

//
//  File enumeration support
//

BOOL
SdirCollectCompressedFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (Opts.GetCompressedFileSize) {
        Entry->CompressedFileSize.LowPart = Opts.GetCompressedFileSize(FullPath, &Entry->CompressedFileSize.HighPart);

        if (Entry->CompressedFileSize.LowPart == INVALID_FILE_SIZE) {
            Entry->CompressedFileSize = Entry->FileSize;
        }
    } else {
        Entry->CompressedFileSize = Entry->FileSize;
    }

    return TRUE;
}

BOOL
SdirCollectCreateTime (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftCreationTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->CreateTime);
    return TRUE;
}

BOOL
SdirCollectFileAttributes (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileAttributes = FindData->dwFileAttributes;
    return TRUE;
}

BOOL
SdirCollectFileId (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->FileId.QuadPart = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->FileId.LowPart = FileInfo.nFileIndexLow;
            Entry->FileId.HighPart = FileInfo.nFileIndexHigh;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectFileName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileNameLengthInChars = (DWORD)_tcslen(FindData->cFileName);
    if (Entry->FileNameLengthInChars >= MAX_PATH - 1) {
        Entry->FileNameLengthInChars = MAX_PATH - 2;
    }
    if (Entry->FileNameLengthInChars > SdirDirCollectionLongest) {
        SdirDirCollectionLongest = Entry->FileNameLengthInChars;
    }

    SdirDirCollectionTotalNameLength += Entry->FileNameLengthInChars;

    memcpy(Entry->FileName, FindData->cFileName, Entry->FileNameLengthInChars * sizeof(TCHAR));
    Entry->FileName[Entry->FileNameLengthInChars + 1] = '\0';

    Entry->Extension = _tcsrchr(Entry->FileName, '.');

    //
    //  For simplicity's sake, if we have no extension set the field
    //  to the end of string, so we'll see a valid pointer of nothing.
    //

    if (Entry->Extension == NULL) {
        Entry->Extension = Entry->FileName + Entry->FileNameLengthInChars;
    } else {
        Entry->Extension++;
    }

    return TRUE;
}

BOOL
SdirCollectFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileSize.LowPart = FindData->nFileSizeLow;
    Entry->FileSize.HighPart = FindData->nFileSizeHigh;
    return TRUE;
}

BOOL
SdirCollectLinkCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->LinkCount = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->LinkCount = FileInfo.nNumberOfLinks;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectOsVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFileRead;
    union {
        IMAGE_DOS_HEADER DosHeader;
        struct {
            DWORD Signature;
            IMAGE_FILE_HEADER ImageHeader;
            IMAGE_OPTIONAL_HEADER OptionalHeader;
        } s2;
    } u1;
    DWORD BytesReturned;

    Entry->OsVersionHigh = 0;
    Entry->OsVersionLow = 0;

    //
    //  We want the earlier handle to be attribute only so we can
    //  operate on directories, but we need data for this, so we
    //  end up with two handles.
    //

    hFileRead = CreateFile(FullPath,
                           FILE_READ_ATTRIBUTES|FILE_READ_DATA,
                           FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_BACKUP_SEMANTICS,
                           NULL);

    if (hFileRead != INVALID_HANDLE_VALUE) {

        if (ReadFile(hFileRead, &u1.DosHeader, sizeof(u1.DosHeader), &BytesReturned, NULL) &&
            BytesReturned == sizeof(u1.DosHeader) &&
            u1.DosHeader.e_magic == IMAGE_DOS_SIGNATURE &&
            u1.DosHeader.e_lfanew != 0) {

            SetFilePointer(hFileRead, u1.DosHeader.e_lfanew, NULL, FILE_BEGIN);

            if (ReadFile(hFileRead, &u1.s2, sizeof(u1.s2), &BytesReturned, NULL) &&
                BytesReturned == sizeof(u1.s2) &&
                u1.s2.Signature == IMAGE_NT_SIGNATURE &&
                u1.s2.ImageHeader.SizeOfOptionalHeader >= FIELD_OFFSET(IMAGE_OPTIONAL_HEADER, Subsystem)) {

                Entry->OsVersionHigh = u1.s2.OptionalHeader.MajorSubsystemVersion;
                Entry->OsVersionLow = u1.s2.OptionalHeader.MinorSubsystemVersion;
            }
        }
        CloseHandle(hFileRead);
    }
    return TRUE;
}

BOOL
SdirCollectReparseTag (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
        Entry->ReparseTag = FindData->dwReserved0;
    } else {
        Entry->ReparseTag = 0;
    }
    return TRUE;
}

BOOL
SdirCollectShortName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    CopyMemory(Entry->ShortFileName, FindData->cAlternateFileName, sizeof(FindData->cAlternateFileName));
    return TRUE;
}

#ifdef UNICODE
BOOL
SdirCollectStreamCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFind;
    SDIR_WIN32_FIND_STREAM_DATA FindStreamData;

    Entry->StreamCount = 0;

    //
    //  These APIs are Unicode only.  We could do an ANSI to Unicode thunk here, but
    //  since Unicode is the default build and ANSI is only useful for older systems
    //  (where this API won't exist) there doesn't seem much point.
    //

    if (Opts.FindFirstStreamW && Opts.FindNextStreamW) {
        hFind = Opts.FindFirstStreamW(FullPath, 0, &FindStreamData, 0);
        if (hFind != INVALID_HANDLE_VALUE) {

            do {
                Entry->StreamCount++;
            } while (Opts.FindNextStreamW(hFind, &FindStreamData));
        }

        FindClose(hFind);
    }

    return TRUE;
}
#endif

BOOL
SdirCollectUsn (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->Usn.QuadPart = 0;
    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {

        struct {
            SDIR_USN_RECORD UsnRecord;
            WCHAR FileName[MAX_PATH];
        } s1;
        DWORD BytesReturned;

        if (DeviceIoControl(hFile, FSCTL_READ_FILE_USN_DATA, NULL, 0, &s1, sizeof(s1), &BytesReturned, NULL)) {
            Entry->Usn.QuadPart = s1.UsnRecord.Usn.QuadPart;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    DWORD Junk;
    PVOID Buffer;
    DWORD VerSize = GetFileVersionInfoSize(FullPath, &Junk);
    VS_FIXEDFILEINFO * RootBlock;

    Entry->FileVersion.QuadPart = 0;
    Entry->FileVersionFlags = 0;

    Buffer = HeapAlloc(GetProcessHeap(), 0, VerSize);
    if (Buffer != NULL) {
        if (GetFileVersionInfo(FullPath, 0, VerSize, Buffer)) {
            if (VerQueryValue(Buffer, _T("\\"), &RootBlock, &Junk)) {
                Entry->FileVersion.HighPart = RootBlock->dwFileVersionMS;
                Entry->FileVersion.LowPart = RootBlock->dwFileVersionLS;
                Entry->FileVersionFlags = RootBlock->dwFileFlags & RootBlock->dwFileFlagsMask;
            }
        }
        HeapFree(GetProcessHeap(), 0, Buffer);
    }
    return TRUE;
}

BOOL
SdirCollectWriteTime(
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftLastWriteTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->WriteTime);
    return TRUE;
}

BOOL
SdirCollectSummary(
    __in PSDIR_DIRENT Entry
    )
{
    if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
        Summary.NumDirs++;
    } else {
        Summary.NumFiles++;
    }

#if SDIR_BIG_SIZE
    Summary.TotalSize += Entry->FileSize.QuadPart;
#else
    Summary.TotalSize += Entry->FileSize.LowPart;
#endif

    if (Opts.EnableCompressedFileSize) {
#if SDIR_BIG_SIZE
        Summary.CompressedSize += Entry->CompressedFileSize.QuadPart;
#else
        Summary.CompressedSize += Entry->CompressedFileSize.LowPart;
#endif
    }

    return TRUE;
}

//
//  Formatting support.  Generic routines used across different types of data.
//

ULONG
SdirDisplayGenericSize(
    __out_ecount(6) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PLARGE_INTEGER GenericSize
    )
{
    TCHAR Suffixes[] = {'b', 'k', 'm', 'g', 't', '?'};
    DWORD SuffixLevel = 0;
    TCHAR StrSize[10];
        
    if (Buffer) {
        LARGE_INTEGER Size = *GenericSize;
        LARGE_INTEGER OldSize = Size;

        while (Size.HighPart != 0 || Size.LowPart > 9999) {
            SuffixLevel++;
            OldSize = Size;

            //
            //  Conceptually we want to divide by 1024.  We do this
            //  in two 32-bit shifts combining back the high bits
            //  so we don't need full compiler support.
            //

            Size.LowPart = (Size.LowPart >> 10) + ((Size.HighPart & 0x3ff) << 22);
            Size.HighPart = Size.HighPart >> 10;
        }

        if (SuffixLevel >= sizeof(Suffixes)/sizeof(TCHAR)) {
            SuffixLevel = sizeof(Suffixes)/sizeof(TCHAR) - 1;
        }
            
        if (Size.LowPart < 100 && SuffixLevel > 0) {
            OldSize.LowPart = (OldSize.LowPart % 1024) * 10 / 1024;
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %2i.%1i%c"),
                        Size.LowPart,
                        OldSize.LowPart,
                        Suffixes[SuffixLevel]);
        } else {
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %4i%c"),
                        Size.LowPart,
                        Suffixes[SuffixLevel]);
        }
            
        SdirPasteStr(Buffer, StrSize, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayHex64 (
    __out_ecount(18) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PLARGE_INTEGER Hex
    )
{
    TCHAR Str[19];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x`%08x"), Hex->HighPart, Hex->LowPart);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 18);
    }

    return 18;
}

ULONG
SdirDisplayHex32 (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in DWORD Hex
    )
{
    TCHAR Str[10];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x"), Hex);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 9);
    }
    return 9;
}

ULONG
SdirDisplayFileDate (
    __out_ecount(11) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrDate[16];

    if (Buffer) {
        _stprintf_s(StrDate,
                    sizeof(StrDate)/sizeof(StrDate[0]),
                    _T(" %04i/%02i/%02i"),
                    Time->wYear,
                    Time->wMonth,
                    Time->wDay);

        SdirPasteStr(Buffer, StrDate, SDIR_FILE_TIME_COLOR, 11);
    }
    return 11;
}

ULONG
SdirDisplayFileTime (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrTime[10];

    if (Buffer) {
        _stprintf_s(StrTime,
                    sizeof(StrTime)/sizeof(StrTime[0]),
                    _T(" %02i:%02i:%02i"),
                    Time->wHour,
                    Time->wMinute,
                    Time->wSecond);

        SdirPasteStr(Buffer, StrTime, SDIR_FILE_TIME_COLOR, 9);
    }
    return 9;
}

//
//  Specific formatting and sizing callbacks for each supported piece of file metadata.
//

ULONG
SdirDisplayCompressedFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayGenericSize(Buffer, SDIR_FILE_SIZE_COLOR, &Entry->CompressedFileSize);
}

ULONG
SdirDisplayFileCreateDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, DefaultAttributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileCreateTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, DefaultAttributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileAttributes (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR StrAtts[32];
    DWORD i;

    if (Buffer) {
        StrAtts[0] = ' ';
        for (i = 0; i < sizeof(AttrPairs)/sizeof(SDIR_ATTRPAIR); i++) {
            if (Entry->FileAttributes & AttrPairs[i].FileAttribute) {
                StrAtts[i + 1] = AttrPairs[i].DisplayLetter;
            } else {
                StrAtts[i + 1] = '-';
            }
        }

        StrAtts[i + 1] = '\0';
    
        SdirPasteStr(Buffer, StrAtts, SDIR_FILE_SIZE_COLOR, sizeof(AttrPairs)/sizeof(SDIR_ATTRPAIR) + 1);
    }
    return sizeof(AttrPairs)/sizeof(SDIR_ATTRPAIR) + 1;
}

ULONG
SdirDisplayFileId (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, DefaultAttributes, &Entry->FileId);
}

ULONG
SdirDisplayFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    if (Buffer) {
        if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
            SdirPasteStr(Buffer, _T(" <LNK>"), SDIR_LNK_COLOR, 6);
            return 6;
        } else if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            SdirPasteStr(Buffer, _T(" <DIR>"), SDIR_DIR_COLOR, 6);
            return 6;
        } else {
            return SdirDisplayGenericSize(Buffer, SDIR_FILE_SIZE_COLOR, &Entry->FileSize);
        }
    }
    return 6;
}

ULONG
SdirDisplayLinkCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->LinkCount);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 4);
    }
    return 4;
}

ULONG
SdirDisplayOsVersion (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[30];
    BYTE  ThisOsMajor = LOBYTE(LOWORD(Opts.OsVersion));
    BYTE  ThisOsMinor = HIBYTE(LOWORD(Opts.OsVersion));
    WORD  Attributes = SDIR_FILE_TIME_COLOR;

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %02i.%02i"), Entry->OsVersionHigh, Entry->OsVersionLow);
    
        if (Entry->OsVersionHigh > ThisOsMajor ||
            Entry->OsVersionHigh == ThisOsMajor && Entry->OsVersionLow > ThisOsMinor) {
    
            Attributes = SDIR_FUTURE_VERSION_COLOR;
        }
    
        SdirPasteStr(Buffer, Str, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayReparseTag (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG Tag = 0;
    if (Entry) {
        Tag = Entry->ReparseTag;
    }
    return SdirDisplayHex32(Buffer, DefaultAttributes, Tag);
}

ULONG
SdirDisplayShortName (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG CurrentChar = 0;
    DWORD ShortLength;

    //
    //  Just because we like special cases, add a space only if both
    //  names are being displayed.
    //

    if (Opts.EnableFileName & SDIR_OPT_DISPLAY) {
        if (Buffer) {
            SdirPasteStrAndPad(&Buffer[CurrentChar], NULL, 0, 0, 1);
        }
        CurrentChar++;
    }

    if (Buffer) {
        ShortLength = (DWORD)_tcslen(Entry->ShortFileName);

        if (ShortLength == 0 && Entry->FileNameLengthInChars <= 12) {
            SdirPasteStrAndPad(&Buffer[CurrentChar], Entry->FileName, DefaultAttributes, Entry->FileNameLengthInChars, 12);
        } else {
            SdirPasteStrAndPad(&Buffer[CurrentChar], Entry->ShortFileName, DefaultAttributes, ShortLength, 12);
        }
    }
    CurrentChar += 12;
    return CurrentChar;
}

ULONG
SdirDisplayStreamCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->StreamCount);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 4);
    }
    return 4;
}

ULONG
SdirDisplayUsn(
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, DefaultAttributes, &Entry->Usn);
}

ULONG
SdirDisplayVersion(
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[40];

    if (Buffer) {
        _stprintf_s(Str,
                    sizeof(Str)/sizeof(Str[0]),
                    _T(" %02i.%02i.%05i.%05i %c%c"),
                    Entry->FileVersion.HighPart >> 16,
                    Entry->FileVersion.HighPart & 0xffff,
                    Entry->FileVersion.LowPart >> 16,
                    Entry->FileVersion.LowPart & 0xffff,
                    Entry->FileVersionFlags & VS_FF_DEBUG ? 'D' : '-',
                    Entry->FileVersionFlags & VS_FF_PRERELEASE ? 'P' : '-');

        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 21);
    }
    return 21;
}

ULONG
SdirDisplaySummary(
    __in WORD DefaultAttributes
    )
{
    SDIR_FMTCHAR Buffer[200];
    TCHAR Str[100];
    DWORD Len;
    LARGE_INTEGER Size;
    ULONG CurrentChar = 0;

    _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %i files, %i dirs,"), Summary.NumFiles, Summary.NumDirs);
    Len = (DWORD)_tcslen(Str);
    SdirPasteStr(&Buffer[CurrentChar], Str, DefaultAttributes, Len);
    CurrentChar += Len;

#if SDIR_BIG_SIZE
    Size.QuadPart = Summary.TotalSize;
#else
    Size.HighPart = 0;
    Size.LowPart = Summary.TotalSize;
#endif

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Size);
    SdirPasteStr(&Buffer[CurrentChar], _T(" used,"), DefaultAttributes, 6);
    CurrentChar += 6;

    if (Opts.EnableCompressedFileSize) {
#if SDIR_BIG_SIZE
        Size.QuadPart = Summary.CompressedSize;
#else
        Size.HighPart = 0;
        Size.LowPart = Summary.CompressedSize;
#endif

        CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Size);
        SdirPasteStr(&Buffer[CurrentChar], _T(" compressed,"), DefaultAttributes, 12);
        CurrentChar += 12;
    }

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Summary.VolumeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol size,"), DefaultAttributes, 10);
    CurrentChar += 10;

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Summary.FreeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol free"), DefaultAttributes, 9);
    CurrentChar += 9;

    SdirWrite(Buffer, CurrentChar);

    return CurrentChar;
}
        

ULONG
SdirDisplayFileWriteDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, DefaultAttributes, &Entry->WriteTime);
}

ULONG
SdirDisplayFileWriteTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, DefaultAttributes, &Entry->WriteTime);
}


//
//  The below table corresponds to the supported options in the program.
//  Each option has a flags value describing whether to collect or display
//  the piece of metadata, a default, a switch to turn display on or off
//  or sort by it, a callback function to tell how much space it will need,
//  an optional sort compare function, and some help text.
//
//  File names and attributes are always collected, because that's how we
//  determine colors.
//


SDIR_OPT SdirOptions[] = {
    {&Opts.EnableCreateDate,         _T("cd"), 0,
        SdirDisplayFileCreateDate,     SdirCollectCreateTime,         SdirCompareCreateTime,         _T("create date")},

    {&Opts.EnableCompressedFileSize, _T("cs"), 0,
        SdirDisplayCompressedFileSize, SdirCollectCompressedFileSize, SdirCompareCompressedFileSize, _T("compressed size")},

    {&Opts.EnableCreateTime,         _T("ct"), 0,
        SdirDisplayFileCreateTime,     SdirCollectCreateTime,         SdirCompareCreateTime,         _T("create time")},

    {&Opts.EnableFileAttributes,     _T("fa"), SDIR_OPT_COLLECT,
        SdirDisplayFileAttributes,     SdirCollectFileAttributes,     NULL,                          _T("file attributes")},

    {&Opts.EnableFileId,             _T("fi"), 0,
        SdirDisplayFileId,             SdirCollectFileId,             SdirCompareFileId,             _T("file id")},

    {&Opts.EnableFileName,           _T("fn"), SDIR_OPT_DISPLAY|SDIR_OPT_COLLECT,
        NULL,                          SdirCollectFileName,           SdirCompareFileName,           _T("file name")},

    {&Opts.EnableFileSize,           _T("fs"), SDIR_OPT_DISPLAY,
        SdirDisplayFileSize,           SdirCollectFileSize,           SdirCompareFileSize,           _T("file size")},

    {&Opts.EnableLinkCount,          _T("lc"), 0,
        SdirDisplayLinkCount,          SdirCollectLinkCount,          SdirCompareLinkCount,          _T("link count")},

    {&Opts.EnableOsVersion,          _T("os"), 0,
        SdirDisplayOsVersion,          SdirCollectOsVersion,          SdirCompareOsVersion,          _T("minimum OS version")},

    {&Opts.EnableReparseTag,         _T("rt"), 0,
        SdirDisplayReparseTag,         SdirCollectReparseTag,         SdirCompareReparseTag,         _T("reparse tag")},

#ifdef UNICODE
    {&Opts.EnableStreamCount,        _T("sc"), 0,
        SdirDisplayStreamCount,        SdirCollectStreamCount,        SdirCompareStreamCount,        _T("stream count")},
#endif

    {&Opts.EnableSummary,            _T("sm"), SDIR_OPT_DISPLAY|SDIR_OPT_COLLECT,
        NULL,                          NULL,                          NULL,                          _T("summary")},

    {&Opts.EnableShortName,          _T("sn"), 0,
        SdirDisplayShortName,          SdirCollectShortName,          SdirCompareShortName,          _T("short name")},

    {&Opts.EnableUsn,                _T("us"), 0,
        SdirDisplayUsn,                SdirCollectUsn,                SdirCompareUsn,                _T("USN")},

    {&Opts.EnableVersion,            _T("vr"), 0,
        SdirDisplayVersion,            SdirCollectVersion,            SdirCompareVersion,            _T("version")},

    {&Opts.EnableWriteDate,          _T("wd"), 0,
        SdirDisplayFileWriteDate,      SdirCollectWriteTime,          SdirCompareWriteTime,          _T("write date")},

    {&Opts.EnableWriteTime,          _T("wt"), 0,
        SdirDisplayFileWriteTime,      SdirCollectWriteTime,          SdirCompareWriteTime,          _T("write time")},
};

//
//  This table is really quite redundant, but it corresponds to the order
//  in which we will render each piece of metadata.  The table above is
//  the order we display help.
//

typedef struct _SDIR_EXEC {
    PUCHAR Enable;
    SDIR_METADATA_WIDTH_FN Function;
} SDIR_EXEC, *PSDIR_EXEC;

SDIR_EXEC SdirExec[] = {
    {&Opts.EnableFileId,             SdirDisplayFileId},
    {&Opts.EnableUsn,                SdirDisplayUsn},
    {&Opts.EnableFileSize,           SdirDisplayFileSize},
    {&Opts.EnableCompressedFileSize, SdirDisplayCompressedFileSize},
    {&Opts.EnableFileAttributes,     SdirDisplayFileAttributes},
    {&Opts.EnableReparseTag,         SdirDisplayReparseTag},
    {&Opts.EnableLinkCount,          SdirDisplayLinkCount},
    {&Opts.EnableStreamCount,        SdirDisplayStreamCount},
    {&Opts.EnableCreateDate,         SdirDisplayFileCreateDate},
    {&Opts.EnableCreateTime,         SdirDisplayFileCreateTime},
    {&Opts.EnableWriteDate,          SdirDisplayFileWriteDate},
    {&Opts.EnableWriteTime,          SdirDisplayFileWriteTime},
    {&Opts.EnableVersion,            SdirDisplayVersion},
    {&Opts.EnableOsVersion,          SdirDisplayOsVersion},
};

BOOL
SdirDisplayCollection();

BOOL
SdirAddToCollection (
    __in PWIN32_FIND_DATA FindData
    ) 
{
    PSDIR_DIRENT CurrentEntry;
    TCHAR FullPath[MAX_PATH];
    DWORD i, j;

    if (SdirDirCollectionCurrent >= SDIR_MAX_DIRENTS) {
        SdirWriteString(_T("Too many files!"));
        return FALSE;
    }

    CurrentEntry = &SdirDirCollection[SdirDirCollectionCurrent];

    SdirDirCollectionCurrent++;

    _stprintf_s(FullPath, MAX_PATH, _T("%s%s"), Opts.ParentName, FindData->cFileName);
    FullPath[MAX_PATH - 1] = '\0';

    //
    //  Copy over the data from Win32's FindFirstFile into our own structure.
    //

    for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {

        //
        //  If we're displaying or sorting, we need the data to use
        //

        if (((*SdirOptions[i].Enable) & SDIR_OPT_COLLECT) &&
               SdirOptions[i].CollectFn) {

            SdirOptions[i].CollectFn(CurrentEntry, FindData, FullPath);
        }
    }

    if (Opts.EnableSummary & SDIR_OPT_COLLECT) {
        SdirCollectSummary(CurrentEntry);
    }

    //
    //  Now that our internal entry is fully poulated, insert it into the correct sorted position.
    //  As an optimization, check if we just need to insert at the end (for file name sort on
    //  NTFS, this is the common case.)
    //

    if (SdirDirCollectionCurrent > 1 &&
        Opts.CompareFn(SdirDirSorted[SdirDirCollectionCurrent - 2], CurrentEntry) == Opts.CompareInverseCondition) {

        SdirDirSorted[SdirDirCollectionCurrent - 1] = CurrentEntry;
        return TRUE;
    }
    for (i = 0; i < SdirDirCollectionCurrent - 1; i++) {
        if (Opts.CompareFn(SdirDirSorted[i], CurrentEntry) == Opts.CompareBreakCondition) {
            for (j = SdirDirCollectionCurrent - 1; j > i; j--) {
                SdirDirSorted[j] = SdirDirSorted[j - 1];
            }
            SdirDirSorted[i] = CurrentEntry;
            return TRUE;
        }
    }
    SdirDirSorted[SdirDirCollectionCurrent - 1] = CurrentEntry;
    return TRUE;
}

BOOL
SdirEnumeratePath (
    __in LPCTSTR FindStr
    )
{
    HANDLE hFind;
    WIN32_FIND_DATA FindData;
    LPTSTR FinalPart;
    DWORD SizeCopied;

    //
    //  At this point we should have a directory and an enumeration criteria.
    //  Copy the directory name and clobber the enumeration criteria.
    //
    
    SizeCopied = GetFullPathName(FindStr, sizeof(Opts.ParentName)/sizeof(Opts.ParentName[0]), Opts.ParentName, &FinalPart);
    if (SizeCopied == 0 || SizeCopied > sizeof(Opts.ParentName)/sizeof(Opts.ParentName[0])) {
        SdirDisplayError(GetLastError(), _T("GetFullPathName"));
        return FALSE;
    }
    Opts.ParentNameLength = SizeCopied;
    *FinalPart = '\0';

    if (Opts.EnableSummary & SDIR_OPT_COLLECT) {

        LARGE_INTEGER Junk;

        //
        //  If GetDiskFreeSpaceEx fails, fall back to GetDiskFreeSpace.  According
        //  to MSDN, this can sometimes be present and fail with not supported.
        //

        if (Opts.GetDiskFreeSpaceEx == NULL ||
            !Opts.GetDiskFreeSpaceEx(Opts.ParentName, &Junk, &Summary.VolumeSize, &Summary.FreeSize)) {

            ULONG BytesPerSector;
            ULONG SectorsPerCluster;
            ULONG FreeClusters;
            ULONG TotalClusters;

            GetDiskFreeSpace(Opts.ParentName, &SectorsPerCluster, &BytesPerSector, &FreeClusters, &TotalClusters);

#if SDIR_BIG_SIZE
            Summary.VolumeSize.QuadPart = SectorsPerCluster * BytesPerSector * TotalClusters;
            Summary.FreeSize.QuadPart = SectorsPerCluster * BytesPerSector * FreeClusters;
#else
            Summary.VolumeSize.LowPart = SectorsPerCluster * BytesPerSector * TotalClusters;
            Summary.FreeSize.LowPart = SectorsPerCluster * BytesPerSector * FreeClusters;
#endif
        }
    }
    
    hFind = FindFirstFile(FindStr, &FindData);
    
    if (hFind == NULL || hFind == INVALID_HANDLE_VALUE) {
        SdirDisplayError(GetLastError(), NULL);
        return FALSE;
    }
    
    do {
        SdirAddToCollection(&FindData);
    } while (FindNextFile(hFind, &FindData) && !Opts.Cancelled);
    
    FindClose(hFind);

    return TRUE;
}


#define SDIR_LINE_ELEMENT_HORIZ      0
#define SDIR_LINE_ELEMENT_TOP_T      1
#define SDIR_LINE_ELEMENT_BOTTOM_T   2
#define SDIR_LINE_ELEMENT_VERT       3

#ifdef UNICODE
TCHAR SdirLineElementsRich[] = {0x2500, 0x252c, 0x2534, 0x2502};
#endif
TCHAR SdirLineElementsText[] = {'-', '+', '+', '|'};


BOOL
SdirDisplayCollection()
{
    PSDIR_DIRENT CurrentEntry;
    DWORD Index, Ext;
    WORD Attributes;
    DWORD Columns;
    DWORD ColumnWidth;
    DWORD ActiveColumn = 0;
    SDIR_FMTCHAR Line[SDIR_MAX_WIDTH];
    DWORD CurrentChar = 0;
    DWORD BufferRows;
    DWORD DisplayedRows = 0;
    DWORD LongestDisplayedFileName = SdirDirCollectionLongest;
    LPTSTR LineElements = SdirLineElementsText;

#ifdef UNICODE
    if (Opts.OutputIsConsole) {
        LineElements = SdirLineElementsRich;
    }
#endif

    //
    //  If we're allowed to shorten names to make the display more
    //  legible, we won't allow a longest name greater than twice
    //  average size.  Since we need to insert three characters to
    //  indicate we butchered a name, don't do this unless we have
    //  a meaningful length to start with (currently 10.)
    //

    if (Opts.EnableNameTruncation) {
        ULONG AverageNameLength;

        AverageNameLength = SdirDirCollectionTotalNameLength / SdirDirCollectionCurrent;
        if (LongestDisplayedFileName > 2 * AverageNameLength && AverageNameLength > 10) {
            LongestDisplayedFileName = 2 * AverageNameLength;
        }
    }

    if (Opts.EnableFileName & SDIR_OPT_DISPLAY) {
        Columns = Opts.ConsoleWidth / (LongestDisplayedFileName + Opts.MetadataWidth);
    } else {
        Columns = Opts.ConsoleWidth / (Opts.MetadataWidth);
    }

    //
    //  If the output is too big to fit the display, force one column
    //  of sufficient size.  The output here will be scrambled, but it
    //  will indicate to the user what happened.
    //

    if (Columns > 0) {
        ColumnWidth = Opts.ConsoleWidth / Columns;
        LongestDisplayedFileName = ColumnWidth - Opts.MetadataWidth;
    } else {
        Columns = 1;
        ColumnWidth = Opts.MetadataWidth;
        if (Opts.EnableFileName & SDIR_OPT_DISPLAY) {
            ColumnWidth += LongestDisplayedFileName;
        }
    }

    //
    //  This really shouldn't happen, even in the worst case of a MAX_PATH name
    //  with all metadata options enabled, but we'll be paranoid.
    //

    if (Columns * ColumnWidth >= SDIR_MAX_WIDTH) {
        SdirWriteString(_T("Too much data for a single line!"));
        return FALSE;
    }

    BufferRows = (SdirDirCollectionCurrent + Columns - 1) / Columns;

    //
    //  Draw the top grid line.
    //

    for (Index = 0; Index < ColumnWidth * Columns; Index++) {
        if (Index % ColumnWidth == ColumnWidth - 1 && Index < (ColumnWidth*Columns-1)) {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_TOP_T];
            Line[Index].Attr = SDIR_GRID_COLOR;
        } else {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_HORIZ];
            Line[Index].Attr = SDIR_GRID_COLOR;
        }
    }

    SdirWrite(Line, Index);
    DisplayedRows++;

    if (ColumnWidth * Columns != Opts.ConsoleWidth || !Opts.OutputIsConsole) {
        SdirWriteString(Opts.NewLine);
    }

    //
    //  Enumerate through the entries.
    //

    for (Index = 0; Index < BufferRows * Columns && !Opts.Cancelled; Index++) {

        //
        //  Because we're sorting down columns first, but rendering a row at a time,
        //  we need to do some matrix math to find which elements belong in which cells.
        //  Some cells in the bottom right might be empty.
        //

        Ext = ActiveColumn * BufferRows + Index / Columns;
        if (Ext < SdirDirCollectionCurrent) {
            CurrentEntry = SdirDirSorted[Ext];
        } else {
            CurrentEntry = NULL;
        }
        Attributes = SDIR_DEFAULT_COLOR;

        //
        //  Render the empty cell.  Or, if we have contents, render that too.
        //

        if (CurrentEntry == NULL) {
            SdirPasteStrAndPad(&Line[CurrentChar], NULL, Attributes, 0, ColumnWidth - 1);
            CurrentChar += ColumnWidth - 1;
        } else {

            //
            //  Determine the color to display each entry from extensions and attributes
            //
    
            for (Ext = 0; Ext < sizeof(SdirExtensionMappings)/sizeof(SDIR_EXTENSION_MAPPING); Ext++) {
                if (_tcsicmp(CurrentEntry->Extension, SdirExtensionMappings[Ext].Extension) == 0) {
                    Attributes = SdirExtensionMappings[Ext].Attributes;
                    break;
                }
            }
    
            //
            //  Kludge special case for directories
            //
    
            if (CurrentEntry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
                Attributes = SDIR_LNK_COLOR;
            } else if (CurrentEntry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                Attributes = SDIR_DIR_COLOR;
            }
    
            //
            //  Paste file name into buffer
            //
    
            if (Opts.EnableFileName & SDIR_OPT_DISPLAY) {
                if (CurrentEntry->FileNameLengthInChars > LongestDisplayedFileName) {
                    ULONG ExtractedLength = (LongestDisplayedFileName - 3) / 2;

                    SdirPasteStr(&Line[CurrentChar], CurrentEntry->FileName, Attributes, ExtractedLength);
                    CurrentChar += ExtractedLength;

                    SdirPasteStr(&Line[CurrentChar], _T("..."), Attributes, ExtractedLength);
                    CurrentChar += 3;

                    SdirPasteStrAndPad(&Line[CurrentChar],
                                       CurrentEntry->FileName + CurrentEntry->FileNameLengthInChars - ExtractedLength,
                                       Attributes,
                                       ExtractedLength,
                                       ColumnWidth - Opts.MetadataWidth - ExtractedLength - 3);

                    CurrentChar += ColumnWidth - Opts.MetadataWidth - ExtractedLength - 3;
                } else {
#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, File name length is capped and used to calculate column width, which is capped.)
                    SdirPasteStrAndPad(&Line[CurrentChar],
                                       CurrentEntry->FileName,
                                       Attributes,
                                       CurrentEntry->FileNameLengthInChars,
                                       ColumnWidth - Opts.MetadataWidth);
                    CurrentChar += ColumnWidth - Opts.MetadataWidth;
                }
            }

            if (Opts.EnableShortName & SDIR_OPT_DISPLAY) {
                CurrentChar += SdirDisplayShortName(&Line[CurrentChar], Attributes, CurrentEntry);

                if (!(Opts.EnableFileName & SDIR_OPT_DISPLAY)) {

                    //
                    //  If file name is hidden, we may need to align things, because
                    //  our columns may contain padding (they're essentially justified.)
                    //

#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, ColumnWidth is maximum file name size (MAX_PATH) plus metadata, padded up to terminal width, which is capped at the size of this buffer.)
                    SdirPasteStrAndPad(&Line[CurrentChar], NULL, 0, 0, ColumnWidth - Opts.MetadataWidth);
                    CurrentChar += ColumnWidth - Opts.MetadataWidth;
                }
            }

            //
            //  Paste any metadata options into the buffer.
            //

            for (Ext = 0; Ext < sizeof(SdirExec)/sizeof(SdirExec[0]); Ext++) {
                if ((*SdirExec[Ext].Enable) & SDIR_OPT_DISPLAY) {
#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, All of this adds up to MetadataWidth, which we sanity checked before)
                    CurrentChar += SdirExec[Ext].Function(&Line[CurrentChar], Attributes, CurrentEntry);
                }
            }
        }
    
        //
        //  We're starting a new column.  If it's the final one we might want a newline,
        //  otherwise we might want a gridline.
        //

        ActiveColumn++;
        if (ActiveColumn % Columns == 0) {

            for (Ext = 0; Opts.NewLine[Ext] != '\0'; Ext++) {
                Line[CurrentChar].Char = Opts.NewLine[Ext];
                Line[CurrentChar].Attr = SDIR_DEFAULT_COLOR;
                CurrentChar++;
            }
            SdirWrite(Line, CurrentChar);
            CurrentChar = 0;
            ActiveColumn = 0;
            DisplayedRows++;

            if (Opts.EnablePause && DisplayedRows >= Opts.ConsoleHeight - 1) {

                if (!SdirPressAnyKey()) {
                    return FALSE;
                }
                DisplayedRows = 0;
            }
        } else {
            Line[CurrentChar].Char = LineElements[SDIR_LINE_ELEMENT_VERT];
            Line[CurrentChar].Attr = SDIR_GRID_COLOR;
            CurrentChar++;
        }
    }

    //
    //  Render the bottom gridline.
    //

    for (Index = 0; Index < ColumnWidth * Columns; Index++) {
        if (Index % ColumnWidth == ColumnWidth - 1 && Index < (ColumnWidth*Columns-1)) {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_BOTTOM_T];
            Line[Index].Attr = SDIR_GRID_COLOR;
        } else {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_HORIZ];
            Line[Index].Attr = SDIR_GRID_COLOR;
        }
    }

    SdirWrite(Line, Index);
    CurrentChar = 0;

    //
    //  We want to restore the 'correct' color back when we return to CMD.
    //

    if (Opts.OutputIsConsole) {
        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), SDIR_DEFAULT_COLOR);
    }

    if (ColumnWidth * Columns != Opts.ConsoleWidth || !Opts.OutputIsConsole) {
        SdirWriteString(Opts.NewLine);
    }

    return TRUE;
}

BOOL WINAPI
SdirRestoreColor (
    __in DWORD dwHandlerType
    )
{
    Opts.Cancelled = TRUE;
    return TRUE;
}

BOOL
SdirInitialize()
{
    CONSOLE_SCREEN_BUFFER_INFO ScreenInfo;
    HANDLE hConsoleOutput = GetStdHandle(STD_OUTPUT_HANDLE);
    HANDLE hKernel32 = GetModuleHandle(_T("KERNEL32"));
    DISABLE_WOW_REDIRECT_FN DisableWow = (DISABLE_WOW_REDIRECT_FN)GetProcAddress(hKernel32, "Wow64DisableWow64FsRedirection");
    ULONG i;

    //
    //  Try to determine the console width so we can size columns.  We
    //  try to use the window size, not the buffer size.
    //  If we're not talking to a console, this might fail, and we have
    //  to fall back to some kind of default.
    //

    if (!GetConsoleScreenBufferInfo(hConsoleOutput, &ScreenInfo)) {
        if (GetLastError() == ERROR_INVALID_HANDLE) {

            Opts.OutputIsConsole = FALSE;
            Opts.NewLine = _T("\r\n");
            if (Opts.ConsoleWidth == 0) {
                Opts.ConsoleWidth = 100;
            }
            Opts.EnablePause = FALSE;

        } else {

            SdirDisplayError(GetLastError(), _T("GetConsoleScreenBufferInfo"));
            return FALSE;
        }

    } else {

        Opts.OutputIsConsole = TRUE;
        Opts.NewLine = _T("\n");

        Opts.ConsoleWidth = ScreenInfo.dwSize.X;
        
        if (Opts.ConsoleWidth > (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1)) {
            Opts.ConsoleWidth = (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1);
        }
    
        Opts.ConsoleHeight = ScreenInfo.dwSize.Y;
        if (Opts.ConsoleHeight > (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1)) {
            Opts.ConsoleHeight = (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1);
        }
    }

    if (Opts.ConsoleWidth > SDIR_MAX_WIDTH) {
        Opts.ConsoleWidth = SDIR_MAX_WIDTH;
    }
    

    //
    //  When running on WOW64, we don't want file system redirection,
    //  because we want people to be able to enumerate those paths
    //

    if (DisableWow) {
        PVOID DontCare;
        DisableWow(&DontCare);
    }

    //
    //  Try to locate FindFirstStream functions if they exist
    //

    Opts.FindFirstStreamW = (FIND_FIRST_STREAM_FN)GetProcAddress(hKernel32, "FindFirstStreamW");
    Opts.FindNextStreamW = (FIND_NEXT_STREAM_FN)GetProcAddress(hKernel32, "FindNextStreamW");
#ifdef UNICODE
    Opts.GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeW");
    Opts.GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExW");
#else
    Opts.GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeA");
    Opts.GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExA");
#endif

    //
    //  Grab the version of the running OS so we can highlight binaries that
    //  need a newer one
    //

    Opts.OsVersion = GetVersion();

    //
    //  Calculate the amount of metadata in each column
    //

    Opts.MetadataWidth = 1;  // Column seperator

    for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {

        //
        //  If we're displaying or sorting, we need the data to use
        //

        if (((*SdirOptions[i].Enable) & SDIR_OPT_DISPLAY) ||
               SdirOptions[i].CompareFn == Opts.CompareFn) {

            *(SdirOptions[i].Enable) |= SDIR_OPT_COLLECT;
        }

        //
        //  If we're displaying, we need space to display in
        //  

        if (((*SdirOptions[i].Enable) & SDIR_OPT_DISPLAY) &&
               SdirOptions[i].WidthFn) {

            Opts.MetadataWidth += SdirOptions[i].WidthFn(NULL, 0, NULL);
        }
    }

    //
    //  Look for Ctrl+C and restore color if we get one.
    //  Not much we can do on failure.
    //

    SetConsoleCtrlHandler(SdirRestoreColor, TRUE);

    SdirDirCollection = HeapAlloc(GetProcessHeap(), 0, SDIR_MAX_DIRENTS * sizeof(SDIR_DIRENT));
    if (SdirDirCollection == NULL) {
        SdirDisplayError(GetLastError(), _T("HeapAlloc"));
        return FALSE;
    }

    SdirDirSorted = HeapAlloc(GetProcessHeap(), 0, SDIR_MAX_DIRENTS * sizeof(PSDIR_DIRENT));
    if (SdirDirSorted == NULL) {
        SdirDisplayError(GetLastError(), _T("HeapAlloc"));
        return FALSE;
    }

    return TRUE;
}


BOOL
SdirUsage()
{
    LPCTSTR strUsage1 = _T("  Usage: [opts] [pathspec] ...\n")
                        _T("\n")
                        _T("   -?           Display usage\n")
                        _T("\n")
                        _T("   -cw[num]     Width of console when writing to files\n")
                        _T("   -p/-pn       Pause/no pause after each screen\n")
                        _T("   -t/-tn       Truncate/no truncate of very long file names\n")
#ifdef UNICODE
                        _T("   -u/-un       Unicode/no unicode output when writing to files\n")
#endif
                        _T("\n");

    LPCTSTR strUsage2 = _T("\n")
                        _T(" Options can also be in the SDIR_OPTS environment variable.\n")
                        _T(" Processed in order, environment then arguments.");

    DWORD  i;
    TCHAR  Line[80];

    _stprintf_s(Line,
                sizeof(Line)/sizeof(Line[0]),
                _T("Sdir version %i.%i, compiled %s, %s, %s"),
                SDIR_VER_MAJOR,
                SDIR_VER_MINOR,
                _T(__DATE__),
#ifdef UNICODE
                _T("Unicode"),
#else
                _T("ANSI"),
#endif
#ifdef MINICRT
                _T("Minicrt")
#else
                _T("Msvcrt")
#endif
                );

    SdirWriteString(Line);

#ifdef _MSC_VER

    _stprintf_s(Line,
                sizeof(Line)/sizeof(Line[0]),
                _T(", C version %i.%02i"),
                _MSC_VER / 100,
                _MSC_VER % 100);

    SdirWriteString(Line);
#endif

    SdirWriteString(_T("\n\n"));

    SdirWriteString(strUsage1);
    for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {

        _stprintf_s(Line,
                    sizeof(Line)/sizeof(Line[0]),
                    _T("   -d%s/-h%s    Display/hide %s%s\n"),
                    SdirOptions[i].Switch,
                    SdirOptions[i].Switch,
                    SdirOptions[i].Help,
                    (SdirOptions[i].Default&SDIR_OPT_DISPLAY)?_T(" (displayed by default)"):_T(""));

        SdirWriteString(Line);

    }
    SdirWriteString(_T("\n"));
    for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
        if (SdirOptions[i].CompareFn) {

            _stprintf_s(Line,
                        sizeof(Line)/sizeof(Line[0]),
                        _T("   -s%s/-i%s    Sort/inverse sort by %s\n"),
                        SdirOptions[i].Switch,
                        SdirOptions[i].Switch,
                        SdirOptions[i].Help);

            SdirWriteString(Line);
        }
    }

    SdirWriteString(strUsage2);
    return TRUE;
}

BOOL
SdirParseOpt (
    __in LPTSTR Opt
    )
{
    BOOL OptParsed = FALSE;
    DWORD i;

    if (Opt[0] == 'd') {
        for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0) {
                *SdirOptions[i].Enable |= SDIR_OPT_DISPLAY;
                OptParsed = TRUE;
            }
        }
    } else if (Opt[0] == 'h') {
        for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0) {
                *SdirOptions[i].Enable &= ~(SDIR_OPT_DISPLAY);
                OptParsed = TRUE;
            }
        }
    } else if (Opt[0] == 's') {
        for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].CompareFn != NULL) {

                Opts.CompareFn = SdirOptions[i].CompareFn;
                Opts.CompareBreakCondition = SDIR_GREATER_THAN;
                Opts.CompareInverseCondition = SDIR_LESS_THAN;
                OptParsed = TRUE;
            }
        }
    } else if (Opt[0] == 'i') {
        for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].CompareFn != NULL) {

                Opts.CompareFn = SdirOptions[i].CompareFn;
                Opts.CompareBreakCondition = SDIR_LESS_THAN;
                Opts.CompareInverseCondition = SDIR_GREATER_THAN;
                OptParsed = TRUE;
            }
        }
    } else if (Opt[0] == 'p') {
        if (Opt[1] == 'n') {
            Opts.EnablePause = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts.EnablePause = TRUE;
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 't') {
        if (Opt[1] == 'n') {
            Opts.EnableNameTruncation = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts.EnableNameTruncation = TRUE;
            OptParsed = TRUE;
        }
#ifdef UNICODE
    } else if (Opt[0] == 'u') {
        if (Opt[1] == 'n') {
            Opts.OutputIsUnicode = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts.OutputIsUnicode = TRUE;
            OptParsed = TRUE;
        }
#endif
    } else if (Opt[0] == 'c') {
        if (Opt[1] == 'w') {
            Opts.ConsoleWidth = _ttoi(&Opt[2]);
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 'r') {

        LPTSTR FinalPart;
        DWORD SizeCopied;
        LPTSTR SubDirPath = &Opt[1];

        OptParsed = TRUE;
        if (SubDirPath[0] == '\0') {
            SubDirPath = _T(".");
        }

        SizeCopied = GetFullPathName(SubDirPath,  sizeof(Opts.SubDirWalk)/sizeof(Opts.SubDirWalk[0]), Opts.SubDirWalk, &FinalPart);
        if (SizeCopied == 0 || SizeCopied >= sizeof(Opts.SubDirWalk)/sizeof(Opts.SubDirWalk[0])) {
            SdirDisplayError(GetLastError(), _T("GetFullPathName"));
            return OptParsed;
        }
    
        Opts.SubDirWalk[sizeof(Opts.SubDirWalk)/sizeof(Opts.SubDirWalk[0]) - 1] = '\0';
    }

    return OptParsed;
}

BOOL
SdirParseArgs (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    ULONG  CurrentArg;
    BOOL   OptParsed;

    TCHAR  EnvOpts[200];
    LPTSTR Opt;
    LPTSTR TokContext = NULL;
    DWORD  i;

    //
    //  Default to name sorting.
    //

    Opts.CompareFn = SdirCompareFileName;
    Opts.CompareBreakCondition = SDIR_GREATER_THAN;
    Opts.CompareInverseCondition = SDIR_LESS_THAN;

    Opts.EnablePause = FALSE;
    Opts.ConsoleWidth = 0;

    Opts.EnableNameTruncation = TRUE;

    Opts.SubDirWalk[0] = '\0';

    //
    //  Default to non-console so errors display until we know what
    //  we're running on
    //

    Opts.OutputIsConsole = FALSE;

    //
    //  Default to ANSI files
    //

    Opts.OutputIsUnicode = FALSE;

    for (i = 0; i < sizeof(SdirOptions)/sizeof(SdirOptions[0]); i++) {
        *SdirOptions[i].Enable = SdirOptions[i].Default;
    }

    if (GetEnvironmentVariable(_T("SDIR_OPTS"), EnvOpts, sizeof(EnvOpts)/sizeof(EnvOpts[0]))) {
        EnvOpts[sizeof(EnvOpts)/sizeof(EnvOpts[0]) - 1] = '\0';
        Opt = _tcstok_s(EnvOpts, _T(" "), &TokContext);

        while (Opt) {

            if (Opt[0] == '-' || Opt[0] == '/') {

                Opt++;

                OptParsed = FALSE;

                OptParsed = SdirParseOpt(Opt);
    
                if (!OptParsed) {
                    SdirWriteString(_T("Unknown environment option: "));
                    SdirWriteString(Opt);
                    SdirWriteString(_T("\n"));
                    SdirUsage();
                    return FALSE;
                }
            }

            Opt = _tcstok_s(NULL, _T(" "), &TokContext);
        }
    }

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++)
    {
        if (argv[CurrentArg][0] == '-' || argv[CurrentArg][0] == '/') {
            LPTSTR Opt = &argv[CurrentArg][1];

            OptParsed = FALSE;

            OptParsed = SdirParseOpt(Opt);

            if (_tcsicmp(Opt, _T("help")) == 0) {
                SdirUsage();
                return FALSE;
            } else if (_tcsicmp(Opt, _T("?")) == 0) {
                SdirUsage();
                return FALSE;
            }

            if (!OptParsed) {
                SdirWriteString(_T("Unknown argument: "));
                SdirWriteString(argv[CurrentArg]);
                SdirWriteString(_T("\n"));
                SdirUsage();
                return FALSE;
            }
        }
    }
    return TRUE;
}


BOOL
SdirEnumerateAndDisplay (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    TCHAR FindStr[MAX_PATH];
    LPTSTR FinalPart;
    DWORD SizeCopied;
    LPTSTR EnumPath;
    ULONG  CurrentArg;
    BOOLEAN EnumerateUserSpecified = FALSE;

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++) {

        if (argv[CurrentArg][0] != '-' && argv[CurrentArg][0] != '/') {

            HANDLE hDir;

            EnumPath = argv[CurrentArg];
            EnumerateUserSpecified = TRUE;

            //
            //  Start by assuming the input needs no processing
            //
    
            SizeCopied = GetFullPathName(EnumPath,  sizeof(FindStr)/sizeof(FindStr[0]), FindStr, &FinalPart);
            if (SizeCopied == 0 || SizeCopied >= sizeof(FindStr)/sizeof(FindStr[0])) {
                SdirDisplayError(GetLastError(), _T("GetFullPathName"));
                return FALSE;
            }
    
            FindStr[sizeof(FindStr)/sizeof(FindStr[0]) - 1] = '\0';
    
            hDir = CreateFile(FindStr, FILE_READ_ATTRIBUTES, FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
    
            if (hDir != INVALID_HANDLE_VALUE) {
                BY_HANDLE_FILE_INFORMATION HandleInfo;
    
                GetFileInformationByHandle(hDir, &HandleInfo);
    
                if (HandleInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                    _stprintf_s(FindStr + SizeCopied, MAX_PATH - SizeCopied, _T("\\*"));
                }
    
                CloseHandle(hDir);
            }

            FindStr[MAX_PATH - 1] = '\0';

            if (!SdirEnumeratePath( FindStr )) {
                return FALSE;
            }

        }
    }

    if (!EnumerateUserSpecified) {
        if (!SdirEnumeratePath( _T("*") )) {
            return FALSE;
        }
    }

    SdirDisplayCollection();

    return TRUE;
}

BOOL
SdirEnumerateAndDisplaySubtree (
    __in TCHAR * TreeRoot,
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    TCHAR NextSubDir[MAX_PATH];
    HANDLE hFind;
    WIN32_FIND_DATA FindData;
    ULONG  CurrentArg;
    BOOLEAN EnumerateUserSpecified = FALSE;

    SdirWriteStringWithAttribute(TreeRoot, SDIR_DIR_COLOR);
    SdirWriteString(_T("\n"));

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++) {

        if (argv[CurrentArg][0] != '-' && argv[CurrentArg][0] != '/') {

            _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\%s"), TreeRoot, argv[CurrentArg]);

            if (NextSubDir[0] == '\0') {
                SdirWriteString(_T("Path exceeds MAX_PATH\n"));
                return FALSE;
            }

            EnumerateUserSpecified = TRUE;

            if (!SdirEnumeratePath( NextSubDir ) && GetLastError() != ERROR_FILE_NOT_FOUND) {
                return FALSE;
            }
        }
    }

    if (!EnumerateUserSpecified) {
        _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\*"), TreeRoot);

        if (NextSubDir[0] == '\0') {
            SdirWriteString(_T("Path exceeds MAX_PATH\n"));
            return FALSE;
        }

        if (!SdirEnumeratePath( NextSubDir ) && GetLastError() != ERROR_FILE_NOT_FOUND) {
            return FALSE;
        }
    }

    if (SdirDirCollectionCurrent > 0) {
        SdirDisplayCollection();
    }

    SdirDirCollectionCurrent = 0;
    SdirDirCollectionLongest = 0;
    SdirDirCollectionTotalNameLength = 0;

    _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\*"), TreeRoot);

    if (NextSubDir[0] == '\0') {
        SdirWriteString(_T("Path exceeds MAX_PATH\n"));
        return FALSE;
    }

    hFind = FindFirstFile(NextSubDir, &FindData);
    
    if (hFind == NULL || hFind == INVALID_HANDLE_VALUE) {
        SdirDisplayError(GetLastError(), NULL);
        return FALSE;
    }
    
    do {
        if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
            _tcscmp(FindData.cFileName, _T(".")) != 0 &&
            _tcscmp(FindData.cFileName, _T("..")) != 0) {

            _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\%s"), TreeRoot, FindData.cFileName);

            if (NextSubDir[0] == '\0') {
                SdirWriteString(_T("Path exceeds MAX_PATH\n"));
                return FALSE;
            }

            if (!SdirEnumerateAndDisplaySubtree( NextSubDir, argc, argv )) {
                FindClose(hFind);
                return FALSE;
            }
            
        }
    } while (FindNextFile(hFind, &FindData) && !Opts.Cancelled);
    
    FindClose(hFind);

    return TRUE;
}


#ifdef UNICODE
#define ENTRYPOINT wmain
#else
#define ENTRYPOINT main
#endif

int __cdecl
ENTRYPOINT (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    if (!SdirParseArgs(argc, argv)) {
        return 1;
    }

    if (!SdirInitialize()) {
        return 1;
    }

    if (Opts.SubDirWalk[0] != '\0') {
        if (!SdirEnumerateAndDisplaySubtree(Opts.SubDirWalk, argc, argv)) {
            return 1;
        }
    } else {
        if (!SdirEnumerateAndDisplay(argc, argv)) {
            return 1;
        }
    }

    if (Opts.EnableSummary & SDIR_OPT_DISPLAY) {

        SdirDisplaySummary(SDIR_DEFAULT_COLOR);
    }

    return 0;
}

// vim:sw=4:ts=4:et:
